# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


import streamlit as st
import pyperclip
import urllib.parse
from urllib.parse import unquote

try:
    import os
    from pathlib import Path
    from dotenv import load_dotenv

    base_dir = os.path.abspath(os.path.dirname(__file__))
    path = Path(base_dir)
    repo_dir = path.parent.parent.absolute()
    load_dotenv(os.path.join(repo_dir, ".env"))
except Exception:
    pass

from prompttools.playground.constants import MODES, MODEL_TYPES, OPENAI_CHAT_MODELS, OPENAI_COMPLETION_MODELS
from prompttools.playground.data_loader import render_prompts, load_data, run_multiple


params = st.experimental_get_query_params()
st.experimental_set_query_params()

st.header("PromptTools Playground")
st.write("Give us a \U00002B50 on [GitHub](https://github.com/hegelai/prompttools)")

with st.sidebar:
    if "mode" not in st.session_state and "mode" in params:
        st.session_state.mode = unquote(params["mode"][0])
    mode = st.radio("Choose a mode", MODES, key="mode")
    if mode != "Model Comparison":
        if "model_type" not in st.session_state and "model_type" in params:
            st.session_state.model_type = unquote(params["model_type"][0])
        model_type = st.selectbox("Model Type", MODEL_TYPES, key="model_type")
        model, api_key = None, None
        if "model" not in st.session_state and "model" in params:
            st.session_state.model = unquote(params["model"][0])
        if model_type in {"LlamaCpp Chat", "LlamaCpp Completion"}:
            model = st.text_input("Local Model Path", key="model")
        elif model_type == "HuggingFace Hub":
            model = st.text_input("Repo ID", key="model")
            api_key = st.text_input("HuggingFace Hub API Key", type="password")
        elif model_type == "Google PaLM":
            model = st.text_input("Model", key="model")
            api_key = st.text_input("Google PaLM API Key", type="password")
        elif model_type == "Anthropic":
            model = st.selectbox("Model", ("claude-2", "claude-instant-1"), key="model")
            api_key = st.text_input("Anthropic API Key", type="password")
        elif model_type == "OpenAI Chat":
            if "model" not in st.session_state and "model" in params:
                st.session_state.model = unquote(params["model"][0])
            model = st.selectbox("Model", OPENAI_CHAT_MODELS, key="model")
            api_key = st.text_input("OpenAI API Key", type="password")
        elif model_type == "OpenAI Completion":
            model = st.selectbox("Model", OPENAI_COMPLETION_MODELS, key="model")
            api_key = st.text_input("OpenAI API Key", type="password")
        elif model_type == "Replicate":
            model = st.text_input("Model", key="model")
            api_key = st.text_input("Replicate API Key", type="password")

        variable_count = 0
        if mode == "Prompt Template":
            instruction_count = st.number_input("Add Template", step=1, min_value=1, max_value=5)
            prompt_count = st.number_input("Add User Input", step=1, min_value=1, max_value=10)
            variable_count = len(params["var_names"][0].split(",")) if "var_names" in params else 1
            variable_count = st.number_input("Add Variable", step=1, min_value=1, max_value=10, value=variable_count)
        elif model_type == "OpenAI Chat":
            instruction_count = st.number_input("Add System Message", step=1, min_value=1, max_value=5)
            prompt_count = st.number_input("Add User Message", step=1, min_value=1, max_value=10)
        else:
            instruction_count = st.number_input("Add Instruction", step=1, min_value=1, max_value=5)
            prompt_count = st.number_input("Add Prompt", step=1, min_value=1, max_value=10)

        var_names = []
        if "var_names" in params:
            var_names = unquote(params["var_names"][0]).split(",")
        for i in range(variable_count):
            if f"varname_{i}" not in st.session_state:
                if len(var_names) > i:
                    st.session_state[f"varname_{i}"] = var_names[i]
                else:
                    st.session_state[f"varname_{i}"] = f"Variable {i+1}"

        for i in range(variable_count):
            var_names.append(
                st.text_input(
                    f"Variable {i+1} Name",
                    key=f"varname_{i}",
                )
            )
        temperature = st.slider("Temperature", min_value=0.01 if model_type == "Replicate" else 0.0, max_value=1.0, value=0.01 if model_type == "Replicate" else 0.0, step=0.01, key="temperature")
        top_p = None
        max_tokens = None
        presence_penalty = None
        frequency_penalty = None
        if model_type == "OpenAI Chat" or model_type == "OpenAI Completion":
            # top_p = st.slider("Top P", min_value=0.0, max_value=1.0, value=1.0, step=0.01, key="top_p")
            # max_tokens = st.number_input("Max Tokens", min_value=0, value=, step=1, key="max_tokens")
            presence_penalty = st.slider(
                "Presence Penalty", min_value=-2.0, max_value=2.0, value=0.0, step=0.01, key="presence_penalty"
            )
            frequency_penalty = st.slider(
                "Frequency Penalty", min_value=-2.0, max_value=2.0, value=0.0, step=0.01, key="frequency_penalty"
            )
    else:
        model_count = st.number_input("Add Model", step=1, min_value=1, max_value=5)
        prompt_count = st.number_input("Add Prompt", step=1, min_value=1, max_value=10)
        openai_api_key = st.text_input("OpenAI API Key", type="password")
        anthropic_api_key = st.text_input("Anthropic Key", type="password")
        google_api_key = st.text_input("Google PaLM Key", type="password")
        hf_api_key = st.text_input("HuggingFace Hub Key", type="password")
        replicate_api_key = st.text_input("Replicate API Key", type="password")


if mode == "Instruction":
    # placeholders = [[st.empty() for _ in range(instruction_count + 1)] for _ in range(prompt_count)]
    placeholders = []

    cols = st.columns(instruction_count + 1)

    with cols[0]:
        a = None
    instructions = []
    for j in range(1, instruction_count + 1):
        if f"instruction_{j-1}" not in st.session_state:
            if "instruction" in params:
                st.session_state[f"instruction_{j-1}"] = unquote(params["instruction"][0])
            else:
                st.session_state[f"instruction_{j-1}"] = "You are a helpful AI assistant."
        with cols[j]:
            instructions.append(
                st.text_area(
                    "System Message" if model_type == "OpenAI Chat" else "Instruction",
                    key=f"instruction_{j-1}",
                )
            )

    prompts = []
    for i in range(prompt_count):
        cols = st.columns(instruction_count + 1)
        if f"prompt_{i}" not in st.session_state and "prompt" in params:
            st.session_state[f"prompt_{i}"] = unquote(params["prompt"][0])
        with cols[0]:
            prompts.append(
                st.text_area(
                    "User Message" if model_type == "OpenAI Chat" else "Prompt",
                    key=f"prompt_{i}",
                )
            )
        placeholders.append([])
        for j in range(1, instruction_count + 1):
            with cols[j]:
                placeholders[i].append(st.empty())  # placeholders for the future output
        st.divider()

    run_button, clear_button, share_button = st.columns([1, 1, 1], gap="small")
    with run_button:
        run = st.button("Run")
    with clear_button:
        clear = st.button("Clear")
    with share_button:
        share = st.button("Share")

    link = "https://prompttools.streamlit.app?"
    link += "mode=" + urllib.parse.quote(mode)
    link += "&model_type=" + urllib.parse.quote(model_type)
    if model:
        link += "&model=" + urllib.parse.quote(model)
    link += "&instruction=" + urllib.parse.quote(instructions[0])
    link += "&prompt=" + urllib.parse.quote(prompts[0])

    if run:
        df = load_data(
            model_type=model_type,
            model=model,
            instructions=instructions,
            user_inputs=prompts,
            temperature=temperature,
            top_p=top_p,
            max_tokens=max_tokens,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            api_key=api_key,
        )
        st.session_state.df = df
        for i in range(len(prompts)):
            for j in range(len(instructions)):
                placeholders[i][j].markdown(df["response"][i + len(prompts) * j])
    elif "df" in st.session_state and not clear:
        df = st.session_state.df
        for i in range(len(prompts)):
            for j in range(len(instructions)):
                placeholders[i][j].markdown(df["response"][i + len(prompts) * j])
elif mode == "Prompt Template":
    instruction = None
    if model_type == "LlamaCpp Chat":
        instruction = st.text_area(
            "Instruction",
            key="instruction",
            value=params["instruction"][0] if "instruction" in params else "You are a helpful AI assistant.",
        )
    elif model_type == "OpenAI Chat":
        instruction = st.text_area(
            "System Message",
            key="instruction",
            value=params["instruction"][0] if "instruction" in params else "You are a helpful AI assistant.",
        )

    placeholders = [[st.empty() for _ in range(instruction_count + variable_count)] for _ in range(prompt_count)]

    cols = st.columns(instruction_count + variable_count)

    with cols[0]:
        a = None
    templates = []
    for j in range(variable_count, instruction_count + variable_count):
        with cols[j]:
            templates.append(
                st.text_area(
                    "Prompt Template",
                    key=f"col_{j-variable_count}",
                    value=params["template"][0] if "template" in params else "",
                )
            )

    vars = []
    varlist = []
    if "vars" in params:
        varlist = params["vars"][0].split(",")
    for i in range(prompt_count):
        cols = st.columns(instruction_count + variable_count)
        vars.append(dict())
        for j in range(variable_count):
            with cols[j]:
                vars[i][var_names[j]] = st.text_area(
                    var_names[j], key=f"var_{i}_{j}", value=varlist[j] if len(varlist) > 0 else ""
                )
        for j in range(variable_count, instruction_count + variable_count):
            with cols[j]:
                placeholders[i][j] = st.empty()  # placeholders for the future output
        st.divider()

    run_button, clear_button, share_button = st.columns([1, 1, 1], gap="small")
    with run_button:
        run = st.button("Run")
    with clear_button:
        clear = st.button("Clear")
    with share_button:
        share = st.button("Share")

    link = "https://prompttools.streamlit.app?"
    link += "mode=" + urllib.parse.quote(mode)
    link += "&model_type=" + urllib.parse.quote(model_type)
    link += "&model=" + urllib.parse.quote(model)
    if instruction:
        link += "&instruction=" + urllib.parse.quote(instruction)
    link += "&template=" + urllib.parse.quote(templates[0])
    if len(var_names) > 0:
        link += "&var_names=" + urllib.parse.quote(",".join(var_names))
    if len(vars) > 0:
        link += "&vars=" + urllib.parse.quote(",".join(vars[0].values()))

    if run:
        prompts = render_prompts(templates, vars)
        df = load_data(model_type, model, [instruction], prompts, temperature, api_key=api_key)
        st.session_state.prompts = prompts
        st.session_state.df = df
        for i in range(len(vars)):
            for j in range(len(templates)):
                placeholders[i][j + variable_count].markdown(df["response"][i + len(vars) * j])
    elif "df" in st.session_state and "prompts" in st.session_state and not clear:
        df = st.session_state.df
        prompts = st.session_state.prompts
        for i in range(len(vars)):
            for j in range(len(templates)):
                placeholders[i][j + variable_count].markdown(df["response"][i + len(vars) * j])
elif mode == "Model Comparison":
    placeholders = [[st.empty() for _ in range(model_count + 1)] for _ in range(prompt_count)]

    cols = st.columns(model_count + 1)

    with cols[0]:
        a = None
    models = []
    model_types = []
    instructions = {}
    for j in range(1, model_count + 1):
        with cols[j]:
            model_types.append(
                st.selectbox(
                    "Model Type",
                    (
                        "OpenAI Chat",
                        "OpenAI Completion",
                        "Anthropic",
                        "Google PaLM",
                        "LlamaCpp Chat",
                        "LlamaCpp Completion",
                        "HuggingFace Hub",
                        "Replicate",
                    ),
                    key=f"type_{j}",
                )
            )
            model = None
            if model_types[j - 1] == "LlamaCpp Chat":
                models.append(st.text_input("Local Model Path", key=f"path_{j}"))
                instructions[j] = st.text_area("Instruction", key=f"instruction_{j}")
            elif model_types[j - 1] == "LlamaCpp Completion":
                models.append(st.text_input("Local Model Path", key=f"path_{j}"))
            elif model_types[j - 1] == "HuggingFace Hub":
                models.append(st.text_input("Repo ID", key=f"model_id_{j}"))
            elif model_types[j - 1] == "Google PaLM":
                models.append(st.text_input("Model", key=f"palm_model_{j}"))
            elif model_types[j - 1] == "Replicate":
                models.append(st.text_input("Model", key=f"replicate_model_{j}"))
                instructions[j] = st.text_area("Instruction", key=f"instruction_{j}")
            elif model_types[j - 1] == "Anthropic":
                models.append(st.selectbox("Model", ("claude-2", "claude-instant-1"), key=f"anthropic_{j}"))
            elif model_types[j - 1] == "OpenAI Chat":
                models.append(
                    st.selectbox(
                        "Model",
                        (
                            "gpt-3.5-turbo",
                            "gpt-3.5-turbo-16k",
                            "gpt-3.5-turbo-0613",
                            "gpt-3.5-turbo-16k-0613",
                            "gpt-3.5-turbo-0301",
                            "gpt-4",
                            "gpt-4-0613",
                            "gpt-4-32k",
                            "gpt-4-32k-0613",
                            "gpt-4-0314",
                            "gpt-4-32k-0314",
                        ),
                        key=f"model_{j}",
                    )
                )
                instructions[j] = st.text_area(
                    "System Message", value="You are a helpful AI assistant.", key=f"instruction_{j}"
                )

    prompts = []
    for i in range(prompt_count):
        cols = st.columns(model_count + 1)
        with cols[0]:
            prompts.append(st.text_area("Prompt", key=f"row_{i}"))
        for j in range(1, model_count + 1):
            with cols[j]:
                placeholders[i][j] = st.empty()  # placeholders for the future output
        st.divider()

    run_button, clear_button, share_button = st.columns([1, 1, 1], gap="small")
    with run_button:
        run = st.button("Run")
    with clear_button:
        clear = st.button("Clear")
    with share_button:
        share = st.button("Share")

    link = "https://prompttools.streamlit.app?"
    link += "mode=" + urllib.parse.quote(mode)

    if run:
        dfs = run_multiple(
            model_types, models, instructions, prompts, openai_api_key, anthropic_api_key, google_api_key, hf_api_key, replicate_api_key
        )
        st.session_state.dfs = dfs
        for i in range(len(prompts)):
            for j in range(len(models)):
                placeholders[i][j + 1].markdown(dfs[j]["response"][i])

    elif "dfs" in st.session_state and not clear:
        dfs = st.session_state.dfs
        for i in range(len(prompts)):
            for j in range(len(models)):
                placeholders[i][j + 1].markdown(dfs[j]["response"][i])

if clear:
    for key in st.session_state.keys():
        del st.session_state[key]


if share:
    try:
        pyperclip.copy(link)
    except pyperclip.PyperclipException:
        st.write("Please copy the following link:")
        st.code(link)



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


import os
import openai
import pandas.core.series
import jinja2
from .error import PromptToolsUtilityError


EVALUATION_SYSTEM_PROMPT = """
Using the provided documents, determine whether or not the response is accurate.
Your answer should be an integer rating from 0 to 10, with 0 being extremely inaccurate
and 10 being perfectly accurate. Only an integer should be returned in the response.
"""

EVALUATION_USER_TEMPLATE = """
DOCUMENTS:
{{documents}}

RESPONSE: {{response}}
ANSWER:
"""


def _get_messages(documents: list[str], response: str):
    environment = jinja2.Environment()
    template = environment.from_string(EVALUATION_USER_TEMPLATE)
    user_message = template.render({"documents": "\n".join(documents), "response": response})
    return [
        {"role": "system", "content": EVALUATION_SYSTEM_PROMPT},
        {"role": "user", "content": user_message},
    ]


def compute(documents: list[str], response: str, model: str = "gpt-4") -> float:
    r"""
    Uses a high quality chat model, like GPT-4, to automatically evaluate a given
    prompt/response pair. Outputs can be 0 or 1.

    Args:
        documents (list[str]): documents to provide relevant context for the model to judge
        model (str): The OpenAI chat model to use for generating an expected response.
            Defaults to GPT-4.
    """
    if not os.environ["OPENAI_API_KEY"]:
        raise PromptToolsUtilityError
    evaluation = openai.chat.completions.create(model=model, messages=_get_messages(documents, response))
    score_text = evaluation.choices[0].message.content
    return int(score_text)


def autoeval_with_documents(
    row: pandas.core.series.Series,
    documents: list[str],
    response_column_name: str = "response",
) -> float:
    r"""
    Given a list of documents, score whether the model response is accurate with "gpt-4" as the judge,
    returning an integer score from 0 to 10.

    Args:
        row (pandas.core.series.Series): A row of data from the full DataFrame (including input, model response, other
            metrics, etc).
        documents (list[str]): documents to provide relevant context for the model to judge
        response_column_name (str): name of the column that contains the model's response, defaults to ``"response"``
    """
    return compute(documents, row[response_column_name])



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

import os
import pandas as pd
from typing import Any, Callable, Dict, Optional

try:
    import chromadb
except ImportError:
    chromadb = None

import itertools
import logging
from prompttools.mock.mock import mock_chromadb_fn
from time import perf_counter
from .experiment import Experiment
from ._utils import _get_dynamic_columns

VALID_TASKS = [""]


class ChromaDBExperiment(Experiment):
    r"""
    Perform an experiment with ``ChromaDB`` to test different embedding functions or retrieval arguments.
    You can query from an existing collection, or create a new one (and insert documents into it) during
    the experiment. If you choose to create a new collection, it will be automatically cleaned up
    as the experiment ends.

    Args:
        chroma_client (chromadb.Client): ChromaDB client to interact with your database
        collection_name (str): the collection that you will get or create
        use_existing_collection (bool): determines whether to create a new collection or use
            an existing one
        query_collection_params (dict[str, list]): parameters used to query the collection
            Each value is expected to be a list to create all possible combinations
        embedding_fns (list[Callable]): embedding functions to test in the experiment
            by default only uses the default one in ChromaDB
        embedding_fn_names (list[str]): names of the embedding functions
        add_to_collection_params (Optional[dict]): documents or embeddings that will be added to
            the newly created collection
    """

    PARAMETER_NAMES = ["chroma_client"]

    def __init__(
        self,
        chroma_client: "chromadb.Client",
        collection_name: str,
        use_existing_collection: bool,
        query_collection_params: dict,
        embedding_fns: list[Callable] = [
            chromadb.utils.embedding_functions.DefaultEmbeddingFunction if chromadb else None
        ],
        embedding_fn_names: list[str] = ["default"],
        add_to_collection_params: Optional[dict] = None,
    ):
        if chromadb is None:
            raise ModuleNotFoundError(
                "Package `chromadb` is required to be installed to use this experiment."
                "Please use `pip install chromadb` to install the package"
            )
        self.chroma_client: "chromadb.Client" = chroma_client
        self.completion_fn = self.chromadb_completion_fn
        if os.getenv("DEBUG", default=False):
            self.completion_fn = mock_chromadb_fn
        self.collection_name = collection_name
        self.use_existing_collection = use_existing_collection
        self.query_collection_params = query_collection_params
        self.embedding_fns = embedding_fns
        self.embedding_fn_names = embedding_fn_names
        if len(embedding_fns) != len(embedding_fn_names):
            raise RuntimeError("Please ensure `embedding_fns` and `embedding_fn_names` are aligned.")
        if use_existing_collection and add_to_collection_params:
            raise RuntimeError("You can either use an existing collection or create a new one during the experiment.")
        if not use_existing_collection and add_to_collection_params is None:
            raise RuntimeError("If you choose to create a new collection, you must also add to it.")
        self.add_to_collection_params = add_to_collection_params if add_to_collection_params else {}
        super().__init__()

    @classmethod
    def initialize(cls, test_parameters: dict[str, list], frozen_parameters: dict):
        required_frozen_params = (
            "chroma_client",
            "collection_name",
            "use_existing_collection",
            "query_collection_params",
        )
        for arg_name in required_frozen_params:
            if arg_name not in frozen_parameters or arg_name in test_parameters:
                raise RuntimeError(f"'{arg_name}' must be a frozen parameter in ChromaDBExperiment.")
        frozen_parameters = {k: [v] for k, v in frozen_parameters.items()}
        return cls(**test_parameters, **frozen_parameters)

    def chromadb_completion_fn(
        self,
        collection: "chromadb.api.Collection",
        **query_params: Dict[str, Any],
    ):
        r"""
        ChromaDB helper function to make request
        """
        results = collection.query(**query_params)
        return results

    def prepare(self) -> None:
        r"""
        Creates argument combinations by taking the cartesian product of all inputs.
        """
        for combo in itertools.product(*self.query_collection_params.values()):
            self.argument_combos.append(dict(zip(self.query_collection_params.keys(), combo)))

    def run(self, runs: int = 1):
        input_args = []  # This will be used to construct DataFrame table
        results = []
        latencies = []
        if not self.argument_combos:
            logging.info("Preparing first...")
            self.prepare()
        for i, emb_fn in enumerate(self.embedding_fns):
            if self.use_existing_collection:
                collection = self.chroma_client.get_collection(self.collection_name, embedding_function=emb_fn)
            else:  # Creating a new collection and add documents
                collection = self.chroma_client.create_collection(self.collection_name, embedding_function=emb_fn)
                collection.add(**self.add_to_collection_params)

            for query_arg_dict in self.argument_combos:
                arg_combo = query_arg_dict.copy()
                arg_combo["embed_fn"] = self.embedding_fn_names[i]  # Save embedding function name to combo
                if "query_texts" in query_arg_dict and "query_embeddings" in query_arg_dict:
                    # `query` does not accept both arguments at the same time
                    continue
                for _ in range(runs):
                    input_args.append(arg_combo)
                    start = perf_counter()

                    results.append(self.chromadb_completion_fn(collection, **query_arg_dict))
                    latencies.append(perf_counter() - start)
            # Clean up
            self.chroma_client.delete_collection(self.collection_name)
        self._construct_result_dfs(input_args, results, latencies)

    def _construct_result_dfs(
        self,
        input_args: list[dict[str, object]],
        results: list[dict[str, object]],
        latencies: list[float],
    ):
        r"""
        Construct a few DataFrames that contain all relevant data (i.e. input arguments, results, evaluation metrics).

        This version only extract the most relevant objects returned by ChromaDB.

        Args:
             input_args (list[dict[str, object]]): list of dictionaries, where each of them is a set of
                input argument that was passed into the model
             results (list[dict[str, object]]): list of responses from the model
             latencies (list[float]): list of latency measurements
        """
        # `input_arg_df` contains all all input args
        input_arg_df = pd.DataFrame(input_args)
        # `dynamic_input_arg_df` contains input args that has more than one unique values
        dynamic_input_arg_df = _get_dynamic_columns(input_arg_df)

        # `response_df` contains the extracted response (often being the text response)
        response_dict = dict()
        response_dict["top doc ids"] = [self._extract_top_doc_ids(result) for result in results]
        response_dict["distances"] = [self._extract_chromadb_dists(result) for result in results]
        response_dict["documents"] = [self._extract_chromadb_docs(result) for result in results]
        response_df = pd.DataFrame(response_dict)
        # `result_df` contains everything returned by the completion function
        result_df = response_df  # pd.concat([self.response_df, pd.DataFrame(results)], axis=1)

        # `score_df` contains computed metrics (e.g. latency, evaluation metrics)
        self.score_df = pd.DataFrame({"latency": latencies})

        # `partial_df` contains some input arguments, extracted responses, and score
        self.partial_df = pd.concat([dynamic_input_arg_df, response_df, self.score_df], axis=1)
        # `full_df` contains all input arguments, responses, and score
        self.full_df = pd.concat([input_arg_df, result_df, self.score_df], axis=1)

    @staticmethod
    def _extract_top_doc_ids(output: Dict[str, object]) -> list[tuple[str, float]]:
        r"""Helper function to get the top document IDs from ChromaDB."""
        return output["ids"][0]

    @staticmethod
    def _extract_chromadb_dists(output: Dict[str, object]) -> list[tuple[str, float]]:
        r"""Helper function to get distances between the prompt and documents from ChromaDB."""
        return output["distances"][0]

    @staticmethod
    def _extract_chromadb_docs(output: Dict[str, object]) -> list[tuple[str, float]]:
        r"""Helper function to get the top documents from ChromaDB."""
        return output["documents"][0]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


def is_interactive() -> bool:
    r"""
    Used to determine if we are in a jupyter notebook, which
    determines how we present the visualizations.
    """
    import __main__ as main

    return not hasattr(main, "__file__")



from unittest import TestCase

from prompttools.experiment import (
    LlamaCppExperiment,
    HuggingFaceHubExperiment,
    OpenAIChatExperiment,
    OpenAICompletionExperiment,
)


class TestExperiment(TestCase):
    # TODO: Currently, it only ensures importing is correct.
    #       Add unit tests to verify initialization.
    def test_llama_cpp_experiment(self):
        pass

    def test_hugging_face_experiment(self):
        pass

    def test_openai_chat_experiment(self):
        pass

    def test_openai_completion_experiment(self):
        pass



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from prompttools.prompttest.threshold_type import ThresholdType


class PromptTestSetupException(Exception):
    r"""
    An exception to throw when something goes wrong with the prompt test setup
    """

    pass


def log_failure(metric_name, threshold, actual, threshold_type):
    r"""
    Prints the test results to the console.
    """
    print(
        "Test failed:  "
        + metric_name
        + "\nThreshold: "
        + (" " * (len("Test failed") - len("Threshold") + 1))
        + str(threshold)
        + "\nActual: "
        + (" " * (len("Test failed") - len("Actual") + 1))
        + str(actual)
        + "\nType: "
        + (" " * (len("Test failed") - len("Type") + 1))
        + str("Minimum" if threshold_type is ThresholdType.MINIMUM else "Maximum")
    )
    print("-" * (len("Test failed: " + metric_name) + 2))



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Any, Dict, List, Union
import itertools
import logging

try:
    from llama_cpp import Llama
except ImportError:
    Llama = None

from time import perf_counter

from .experiment import Experiment
from .error import PromptExperimentException
from prompttools.selector.prompt_selector import PromptSelector


class LlamaCppExperiment(Experiment):
    r"""
    Used to experiment across parameters for a local model, supported by LlamaCpp and GGML.

    Note:
        - All arguments here should be a ``list``, even if you want to keep the argument frozen
          (i.e. ``temperature=[1.0]``), because the experiment will try all possible combination
          of the input arguments. For example, ``model_params`` should have string keys,
          with ``list``s being the values.

    Args:
        model_path (List[str]): list of paths to the models that you would like to run
        prompt (List[str] | List[PromptSelector]): list of prompts to test
        model_params (Dict[str, list[object]]): Parameters for initializing the model. The values should be ``list``s.
        call_params: (Dict[str, list[object]]): Parameters for calling the model completion function.
            The values should be ``list``s.
    """

    MODEL_PARAMETERS = (
        "model_path",
        "lora_path",
        "lora_base",
        "n_ctx",
        "n_parts",
        "seed",
        "f16_kv",
        "logits_all",
        "vocab_only",
        "use_mlock",
        "n_threads",
        "n_batch",
        "use_mmap",
        "last_n_tokens_size",
        "verbose",
    )

    CALL_PARAMETERS = (
        "prompt",
        "suffix",
        "max_tokens",
        "temperature",
        "top_p",
        "logprobs",
        "echo",
        "stop",
        "repeat_penalty",
        "top_k",
    )

    DEFAULT = {
        "lora_path": [None],
        "lora_base": [None],
        "n_ctx": [512],
        "n_parts": [-1],
        "seed": [1337],
        "f16_kv": [True],
        "logits_all": [False],
        "vocab_only": [False],
        "use_mlock": [False],
        "n_threads": [None],
        "n_batch": [512],
        "use_mmap": [True],
        "last_n_tokens_size": [64],
        "verbose": [True],
        "suffix": [None],
        "max_tokens": [128],
        "temperature": [0.8],
        "top_p": [0.95],
        "logprobs": [None],
        "echo": [False],
        "stop": [None],
        "repeat_penalty": [1.1],
        "top_k": [40],
    }

    def __init__(
        self,
        model_path: List[str],
        prompt: Union[List[str], List[PromptSelector]],
        model_params: Dict[str, list[object]] = {},
        call_params: Dict[str, list[object]] = {},
    ):
        if Llama is None:
            raise ModuleNotFoundError(
                "Package `llama-cpp-python` is required to be installed to use this experiment."
                "Please use `pip install llama-cpp-python` to install the package"
            )
        self.completion_fn = self.llama_completion_fn
        self.model_params = model_params
        self.call_params = call_params
        self.model_params["model_path"] = model_path

        # If we are using a prompt selector, we need to
        # render the prompts from the selector
        if isinstance(prompt[0], PromptSelector):
            self.call_params["prompt"] = [selector.for_llama() for selector in prompt]
        else:
            self.call_params["prompt"] = prompt

        # Set defaults
        for param in self.MODEL_PARAMETERS:
            if param not in self.model_params:
                self.model_params[param] = self.DEFAULT[param]
        for param in self.CALL_PARAMETERS:
            if param not in self.call_params:
                self.call_params[param] = self.DEFAULT[param]
        self.all_args = self.model_params | self.call_params
        super().__init__()

    @classmethod
    def initialize(cls, test_parameters: dict[str, list], frozen_parameters: dict):
        raise NotImplementedError(
            "`initialize` is currently not compatible with `LlamaCppExperiment`. Please" "use `__init__` directly."
        )

    def prepare(self) -> None:
        r"""
        Creates argument combinations by taking the cartesian product of all inputs.
        """
        self.model_argument_combos = [
            dict(zip(self.model_params, val)) for val in itertools.product(*self.model_params.values())
        ]
        self.call_argument_combos = [
            dict(zip(self.call_params, val)) for val in itertools.product(*self.call_params.values())
        ]

    def llama_completion_fn(
        self,
        **params: Dict[str, Any],
    ):
        r"""
        Local model helper function to make request
        """
        client = params["client"]
        call_params = {k: v for k, v in params.items() if k != "client"}
        response = client(**call_params)
        return response

    def run(
        self,
        runs: int = 1,
    ) -> None:
        r"""
        Create tuples of input and output for every possible combination of arguments.
        For each combination, it will execute `runs` times, default to 1.
        For local models we need to run this in a single thread.
        """
        if not self.argument_combos:
            logging.info("Preparing first...")
            self.prepare()
        results = []
        latencies = []
        for model_combo in self.model_argument_combos:
            client = Llama(**model_combo)
            for call_combo in self.call_argument_combos:
                for _ in range(runs):
                    call_combo["client"] = client
                    start = perf_counter()
                    res = self.completion_fn(**call_combo)
                    latencies.append(perf_counter() - start)
                    results.append(res)
                    self.argument_combos.append(model_combo | call_combo)
        if len(results) == 0:
            logging.error("No results. Something went wrong.")
            raise PromptExperimentException
        self._construct_result_dfs(self.argument_combos, results, latencies, extract_response_equal_full_result=True)

    @staticmethod
    def _extract_responses(output: Dict[str, object]) -> list[str]:
        return [choice["text"] for choice in output["choices"]][0]

    def _get_model_names(self):
        return [os.path.basename(combo["model_path"]) for combo in self.argument_combos]

    def _get_prompts(self):
        return [combo["prompt"] for combo in self.argument_combos]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


from . import autoeval, expected, validate_json, validate_python, similarity
from .autoeval import autoeval_binary_scoring
from .autoeval_from_expected import autoeval_from_expected_response
from .autoeval_scoring import autoeval_scoring
from .autoeval_with_docs import autoeval_with_documents
from .chunk_text import chunk_text
from .expected import compute_similarity_against_model
from .ranking_correlation import ranking_correlation
from .similarity import semantic_similarity
from .validate_json import validate_json_response
from .validate_python import validate_python_response

__all__ = [
    "autoeval",
    "autoeval_binary_scoring",
    "autoeval_from_expected_response",
    "autoeval_scoring",
    "autoeval_with_documents",
    "chunk_text",
    "compute_similarity_against_model",
    "expected",
    "validate_json",
    "validate_json_response",
    "validate_python",
    "validate_python_response",
    "ranking_correlation",
    "semantic_similarity",
    "similarity",
]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Any, Dict, List
import itertools

from time import perf_counter
import logging

try:
    from langchain.chains import LLMChain, SequentialChain
    from langchain.chains.router.llm_router import LLMRouterChain, RouterOutputParser
    from langchain.chains.router.multi_prompt_prompt import MULTI_PROMPT_ROUTER_TEMPLATE
    from langchain.prompts import PromptTemplate
    from langchain.chains.router import MultiPromptChain
    from langchain.chains import ConversationChain
except ImportError:
    LLMChain = None

from prompttools.mock.mock import mock_lc_completion_fn

from .experiment import Experiment
from .error import PromptExperimentException


class SequentialChainExperiment(Experiment):
    r"""
    Experiment for testing LangChain's sequential chains.

    Args:
        llm (list): list of LLMs
        prompt_template (list[list]): list of prompt templates
        prompt (list[str]): list of prompts
        kwargs (dict): keyword arguments to call the model with
    """

    MODEL_PARAMETERS = ["llm"]

    CALL_PARAMETERS = ["prompt_template", "prompt"]

    def __init__(
        self,
        llm: List[Any],
        prompt_template: List[List[Any]],
        prompt: List[str],
        **kwargs: Dict[str, object],
    ):
        if LLMChain is None:
            raise ModuleNotFoundError(
                "Package `langchain` is required to be installed to use this experiment."
                "Please use `pip install langchain` to install the package"
            )
        self.completion_fn = self.lc_completion_fn
        if os.getenv("DEBUG", default=False):
            self.completion_fn = mock_lc_completion_fn
        self.model_params = dict(llm=llm)  # placeholder for future

        self.call_params = dict(prompt_template=prompt_template, prompt=prompt)
        for k, v in kwargs.items():
            self.CALL_PARAMETERS.append(k)
            self.call_params[k] = v

        self.all_args = self.model_params | self.call_params
        super().__init__()

    def prepare(self) -> None:
        r"""
        Combo builder.
        """
        self.model_argument_combos = [
            dict(zip(self.model_params, val, strict=False)) for val in itertools.product(*self.model_params.values())
        ]
        self.call_argument_combos = [
            dict(zip(self.call_params, val, strict=False)) for val in itertools.product(*self.call_params.values())
        ]

    def lc_completion_fn(
        self,
        **params: Dict[str, Any],
    ):
        r"""
        Local model helper function to make request.
        """
        client = params["client"]
        response = client(params["prompt"])
        return response

    def run(
        self,
        runs: int = 1,
    ) -> None:
        r"""
        Create tuples of input and output for every possible combination of arguments.
        For each combination, it will execute `runs` times, default to 1.
        # TODO This can be done with an async queue.
        """
        if not self.argument_combos:
            logging.info("Preparing first...")
            self.prepare()
        results = []
        latencies = []
        for model_combo in self.model_argument_combos:
            for call_combo in self.call_argument_combos:
                llm = model_combo["llm"]
                llm = llm(temperature=call_combo["temperature"])
                chain = []
                for i, prompt_template in enumerate(call_combo["prompt_template"]):
                    chain.append(LLMChain(llm=llm, prompt=prompt_template, output_key=call_combo["output_key"][i]))
                client = SequentialChain(
                    chains=chain,
                    input_variables=call_combo["input_variables"],
                    output_variables=call_combo["output_variables"],
                    verbose=True,
                )
                for _ in range(runs):
                    call_combo["client"] = client
                    start = perf_counter()
                    res = self.completion_fn(**call_combo)
                    latencies.append(perf_counter() - start)
                    results.append(res)
                    self.argument_combos.append(model_combo | call_combo)
        if len(results) == 0:
            logging.error("No results. Something went wrong.")
            raise PromptExperimentException
        self._construct_result_dfs(self.argument_combos, results, latencies, extract_response_equal_full_result=True)

    @staticmethod
    def _extract_responses(output: List[Dict[str, object]]) -> str:
        return str({k: output[k] for k in output})


class RouterChainExperiment(Experiment):
    r"""
    Experiment for testing LangChain's router chains.

    Args:
        llm (list): list of LLMs
        prompt_infos (list(list[dict(Any)])): list of list of dicts describing key features of prompt chain
        prompt (list[str]): list of prompts
        kwargs (dict): keyword arguments to call the model with
    """

    MODEL_PARAMETERS = ["llm"]

    CALL_PARAMETERS = ["prompt_infos", "prompt"]

    def __init__(
        self,
        llm: List[Any],
        prompt_infos: List[List[Dict[str, Any]]],
        prompt: List[str],
        **kwargs: Dict[str, object],
    ):
        if LLMChain is None:
            raise ModuleNotFoundError(
                "Package `langchain` is required to be installed to use this experiment."
                "Please use `pip install langchain` to install the package"
            )
        self.completion_fn = self.lc_completion_fn
        if os.getenv("DEBUG", default=False):
            self.completion_fn = mock_lc_completion_fn
        self.model_params = dict(llm=llm)  # placeholder for future

        self.call_params = dict(prompt_infos=prompt_infos, prompt=prompt)
        for k, v in kwargs.items():
            self.CALL_PARAMETERS.append(k)
            self.call_params[k] = v

        self.all_args = self.model_params | self.call_params
        super().__init__()

    def prepare(self) -> None:
        r"""
        Combo builder.
        """
        self.model_argument_combos = [
            dict(zip(self.model_params, val, strict=False)) for val in itertools.product(*self.model_params.values())
        ]
        self.call_argument_combos = [
            dict(zip(self.call_params, val, strict=False)) for val in itertools.product(*self.call_params.values())
        ]

    def lc_completion_fn(
        self,
        **params: Dict[str, Any],
    ):
        r"""
        Local model helper function to make request.
        """
        client = params["client"]
        response = client.run(params["prompt"])
        return response

    def run(
        self,
        runs: int = 1,
    ) -> None:
        r"""
        Create tuples of input and output for every possible combination of arguments.
        For each combination, it will execute `runs` times, default to 1.
        # TODO This can be done with an async queue.
        """
        if not self.argument_combos:
            logging.info("Preparing first...")
            self.prepare()
        results = []
        latencies = []
        for model_combo in self.model_argument_combos:
            for call_combo in self.call_argument_combos:
                llm = model_combo["llm"]
                llm = llm(temperature=call_combo["temperature"])

                destination_chains = {}
                prompt_infos = call_combo["prompt_infos"]
                for p_info in prompt_infos:
                    name = p_info["name"]
                    prompt_template = p_info["prompt_template"]
                    prompt = PromptTemplate(template=prompt_template, input_variables=["input"])
                    chain = LLMChain(llm=llm, prompt=prompt)
                    destination_chains[name] = chain
                default_chain = ConversationChain(llm=llm, output_key="text")

                destinations = [f"{p['name']}: {p['description']}" for p in prompt_infos]
                destinations_str = "\n".join(destinations)
                router_template = MULTI_PROMPT_ROUTER_TEMPLATE.format(destinations=destinations_str)
                router_prompt = PromptTemplate(
                    template=router_template,
                    input_variables=["input"],
                    output_parser=RouterOutputParser(),
                )
                router_chain = LLMRouterChain.from_llm(llm, router_prompt)
                client = MultiPromptChain(
                    router_chain=router_chain,
                    destination_chains=destination_chains,
                    default_chain=default_chain,
                    verbose=True,
                )
                for _ in range(runs):
                    call_combo["client"] = client
                    start = perf_counter()
                    res = self.completion_fn(**call_combo)
                    latencies.append(perf_counter() - start)
                    results.append(res)
                    call_combo["client"] = llm  # RouterOutputParser not equipped for serialization
                    self.argument_combos.append(model_combo | call_combo)
        if len(results) == 0:
            logging.error("No results. Something went wrong.")
            raise PromptExperimentException
        self._construct_result_dfs(self.argument_combos, results, latencies, extract_response_equal_full_result=True)

    @staticmethod
    def _extract_responses(output: List[Dict[str, object]]) -> str:
        return output



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


import os
import pandas.core.series
import jinja2

try:
    import anthropic
except ImportError:
    anthropic = None


AUTO_EVAL_PROMPT_TEMPLATE = """
{{HUMAN_PROMPT}} Given the fact {{fact}}

Evaluate the following Answer on a scale from 1 - 7. Please only respond with an integer from 1 - 7 with no other text.
Lower score means the answer is factually wrong, higher score means the answer is correct. A medium score for
uncertain but not wrong.

Answer: {{model_answer}}

{{AI_PROMPT}}
"""


def _generate_auto_eval_prompt(fact: str, model_answer: str):
    environment = jinja2.Environment()
    template = environment.from_string(AUTO_EVAL_PROMPT_TEMPLATE)
    auto_eval_prompt = template.render(
        {
            "HUMAN_PROMPT": anthropic.HUMAN_PROMPT,
            "AI_PROMPT": anthropic.AI_PROMPT,
            "fact": fact,
            "model_answer": model_answer,
        }
    )
    return auto_eval_prompt


def compute(fact: str, model_answer: str, model: str = "claude-2") -> float:
    r"""
    Uses a high quality chat model, like claude-2, to automatically score a given
    fact/response pair. Output should be an integer ranging from 1 - 7.

    Args:
        fact (str): The fact (truth). The auto-eval model will judge how close the ``response`` is
            from this fact (truth).
        model_answer (str): The model response.
        model (str): The model that will be judging how close is the response from the truth.
            Defaults to Claude 2.
    """
    if not os.environ["ANTHROPIC_API_KEY"]:
        raise RuntimeError("Missing API key for evaluation.")
    client = anthropic.Anthropic(api_key=os.environ["ANTHROPIC_API_KEY"])
    completion_response = client.completions.create(
        max_tokens_to_sample=100, model=model, prompt=_generate_auto_eval_prompt(fact, model_answer)
    )
    return int(completion_response.completion)


def autoeval_scoring(row: pandas.core.series.Series, expected: str, response_column_name: str = "response") -> float:
    r"""
    Uses auto-evaluation to score the model response.

    Args:
        row (pandas.core.series.Series): A row of data from the full DataFrame (including input, model response, other
            metrics, etc).
        expected (str): the expected response
        response_column_name (str): name of the column that contains the model's response, defaults to ``"response"``
    """
    if anthropic is None:
        raise ModuleNotFoundError(
            "Package `anthropic` is required to be installed to use this experiment."
            "Please use `pip install anthropic` to install the package"
        )
    return compute(fact=expected, model_answer=row[response_column_name])



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

import logging
import itertools
from functools import partial

from prompttools.mock.mock import mock_replicate_stable_diffusion_completion_fn
from IPython.display import display, HTML
from tabulate import tabulate

from prompttools.selector.prompt_selector import PromptSelector
from ..widgets.utility import is_interactive


import os


try:
    import replicate

except ImportError:
    replicate = None

from .experiment import Experiment


class ReplicateExperiment(Experiment):
    r"""
    Perform an experiment with the Replicate API for both image models and LLMs.

    Note:
        Set your API token to ``os.environ["REPLICATE_API_TOKEN"]``.
        If you are using an image model, set ``use_image_model=True`` as input argument.

    Args:
        models (list[str]): "stability-ai/stable-diffusion:27b93a2413e"
        input_kwargs (dict[str, list]): keyword arguments that can be used across all models
        model_specific_kwargs (dict[str, dict[str, list]]): model-specific keyword arguments that will only be used
            by a specific model (e.g. ``stability-ai/stable-diffusion:27b93a2413``
        use_image_model (bool): Defaults to ``False``, must set to ``True`` to render output from image models.
    """

    def __init__(
        self,
        models: list[str],
        input_kwargs: dict[str, list],
        model_specific_kwargs: dict[str, dict[str, list]] = {},
        use_image_model: bool = False,
    ):
        if replicate is None:
            raise ModuleNotFoundError(
                "Package `replicate` is required to be installed to use this experiment."
                "Please use `pip install replicate` to install the package"
            )
        try:
            os.environ["REPLICATE_API_TOKEN"]
        except KeyError:
            raise RuntimeError('`os.environ["REPLICATE_API_TOKEN]` needs to be set.')
        
        # If we are using a prompt selector, we need to
        # render the prompts from the selector        
        if isinstance(input_kwargs['prompt'][0], PromptSelector):
            input_kwargs['prompt'] = [selector.for_llama() for selector in input_kwargs['prompt']]

        self.models = models
        self.input_kwargs = input_kwargs
        self.model_specific_kwargs = model_specific_kwargs
        if os.getenv("DEBUG", default=False):
            self.completion_fn = mock_replicate_stable_diffusion_completion_fn
        else:
            self.completion_fn = self.replicate_completion_fn
        super().__init__()
        self.image_experiment = use_image_model

    def prepare(self):
        for model in self.models:
            for base_combo in itertools.product(*self.input_kwargs.values()):
                arg_dict = dict(zip(self.input_kwargs.keys(), base_combo))
                # arg_dict['model_version'] = model
                for model_combo in itertools.product(*self.model_specific_kwargs[model].values()):
                    model_arg_dict = dict(zip(self.model_specific_kwargs[model].keys(), model_combo))
                    for k, v in model_arg_dict.items():
                        arg_dict[k] = v
                    arg_dict["model_version"] = model
                    self.argument_combos.append(arg_dict)

    @staticmethod
    def replicate_completion_fn(model_version: str, **kwargs):
        return replicate.run(model_version, input=kwargs)

    def _extract_responses(self, output) -> str:
        if self.image_experiment:
            return output[0]  # Output should be a list of URIs
        else:  # Assume `output` is a generator of text
            res = ""
            for item in output:
                res += item
            return res

    @staticmethod
    def _image_tag(url, image_width):
        r"""
        Create the HTML code to render the image.
        """
        return f'<img src="{url}" width="{image_width}"/>'

    def visualize(
        self, get_all_cols: bool = False, pivot: bool = False, pivot_columns: list = [], image_width: int = 300
    ) -> None:
        if not self.image_experiment:
            super().visualize(get_all_cols, pivot, pivot_columns)
        else:
            if pivot:
                table = self.pivot_table(pivot_columns, get_all_cols=get_all_cols)
            else:
                table = self.get_table(get_all_cols)

            images = table["response"].apply(partial(self._image_tag, image_width=image_width))
            table["images"] = images

            if is_interactive():
                display(HTML(table.to_html(escape=False, columns=[col for col in table.columns if col != "response"])))
            else:
                logging.getLogger().setLevel(logging.INFO)
                logging.info(tabulate(table, headers="keys", tablefmt="psql"))



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


from typing import Callable, Dict, List
from collections import defaultdict
from prompttools.experiment import Experiment
import pandas as pd


class MultiExperimentHarness:
    r"""
    This is designed to run experiments across multiple model providers. The underlying APIs for different models
    (e.g. LlamaCpp and OpenAI) are different, this provides a way to manage that complexity.
    This will run experiments for different providers, and combine the results into a single table.

    The notebook "examples/notebooks/GPT4vsLlama2.ipynb" provides a good example how this can used
    to test prompts across different models.

    Args:
        experiments (list[Experiment]): The list of experiments that you would like to execute (e.g.
            ``prompttools.experiment.OpenAICompletionExperiment``)
    """

    def __init__(self, experiments: List[Experiment]):
        self.experiments = experiments

    def prepare(self):
        for experiment in self.experiments:
            experiment.prepare()

    def run(self):
        for experiment in self.experiments:
            experiment.run()

    def evaluate(self, metric_name: str, eval_fn: Callable) -> None:
        for experiment in self.experiments:
            experiment.evaluate(metric_name, eval_fn)

    def gather_feedback(self) -> None:
        pass

    def _get_argument_combos(self):
        tmp = [combo for experiment in self.experiments for combo in experiment.argument_combos]
        return tmp

    def _get_prompts(self):
        tmp = [combo for experiment in self.experiments for combo in experiment._get_prompts()]
        return tmp

    def _get_results(self):
        tmp = [
            experiment._extract_responses(result) for experiment in self.experiments for result in experiment.results
        ]
        return tmp

    def _get_scores(self):
        scores = defaultdict(list)
        for experiment in self.experiments:
            for name, score in experiment.scores.items():
                scores[name].extend(score)
        return scores

    def _get_experiment_names(self):
        tmp = [name for experiment in self.experiments for name in experiment._get_model_names()]
        return tmp

    def visualize(self, colname: str = None) -> None:
        scores = self._get_scores()
        data = {
            "prompt": self._get_prompts(),
            "response(s)": self._get_results(),
            "latency": scores["latency"],
            "model": self._get_experiment_names(),
        }
        # Add scores for each eval fn, including feedback
        for metric_name, evals in scores.items():
            if metric_name != "comparison":
                data[metric_name] = evals
        df = pd.DataFrame(data)
        if colname:
            df = pd.pivot_table(
                df,
                values=colname,
                index=["prompt"],
                columns=["model"],
                aggfunc=lambda x: x.iloc[0],
            )
        return df

    def rank(self, metric_name: str, is_average: bool = False) -> Dict[str, float]:
        pass



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.
import json

import requests
import threading
import queue
from functools import partial
import openai
import os
from dotenv import load_dotenv
from os.path import join, dirname
from time import perf_counter
from prompttools.common import HEGEL_BACKEND_URL


# Load "OPENAI_API_KEY" into `os.environ["OPENAI_API_KEY"]`
# See `.env.example`
dotenv_path = join(dirname(__file__), ".env")
load_dotenv(dotenv_path)


class Logger:
    def __init__(self):
        self.backend_url = f"{HEGEL_BACKEND_URL}/sdk/logger"
        self.data_queue = queue.Queue()
        self.worker_thread = threading.Thread(target=self.worker)

        # When the main thread is joining, put `None` into queue to signal worker thread to end
        threading.Thread(target=lambda: threading.main_thread().join() or self.data_queue.put(None)).start()

        self.worker_thread.start()

    def execute_and_add_to_queue(self, callable_func, **kwargs):
        if "hegel_model" in kwargs:
            hegel_model = kwargs["hegel_model"]
            del kwargs["hegel_model"]
        else:
            hegel_model = None
        start = perf_counter()
        result = callable_func(**kwargs)
        latency = perf_counter() - start
        self.data_queue.put(
            {
                "hegel_model": hegel_model,
                "result": result.model_dump_json(),
                "input_parameters": json.dumps(kwargs),
                "latency": latency,
            }
        )
        return result

    def wrap(self, callable_func):
        return partial(self.execute_and_add_to_queue, callable_func)

    def worker(self):
        while True:
            if not self.data_queue.empty():
                result = self.data_queue.get()
                if result is None:
                    return
                self.log_data_to_remote(result)
                self.data_queue.task_done()

    def log_data_to_remote(self, data):
        try:
            headers = {
                "Content-Type": "application/json",
                "Authorization": os.environ["HEGELAI_API_KEY"],
            }

            response = requests.post(self.backend_url, json=data, headers=headers)
            if response.status_code != 200:
                print(f"Failed to send data to Flask API. Status code: {response.status_code} for {data}.")
        except requests.exceptions.RequestException as e:
            print(f"Error sending data to Flask API: {e}")


sender = Logger()
# Monkey-patching
try:
    openai.chat.completions.create = sender.wrap(openai.chat.completions.create)
except Exception:
    print("You may need to add `OPENAI_API_KEY=''` to your `.env` file.")
    raise



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


import os
import openai
import pandas.core.series
from .error import PromptToolsUtilityError
from . import similarity


def compute(prompt: str, model: str = "gpt-4") -> str:
    r"""
    Computes the expected result of a given prompt by using a high
    quality LLM, like GPT-4.

    Args:
        prompt (str): The input prompt.
        model (str): The OpenAI chat model to use for generating an expected response.
            Defaults to GPT-4.
    """
    if not os.environ["OPENAI_API_KEY"]:
        raise PromptToolsUtilityError
    response = openai.chat.completions.create(
        model=model,
        messages=[
            {"role": "user", "content": prompt},
        ],
    )
    return response.choices[0].message.content


def evaluate(prompt: str, response: str, model: str = "gpt-4") -> str:
    r"""
    Computes the similarity of a given response to the expected result
    generated from a high quality LLM (by default GPT-4) using the same prompt.

    Args:
        prompt (str): The input prompt.
        response (str): The model response.
        model (str): The OpenAI chat model to use for generating an expected response.
            Defaults to GPT-4.
    """
    expected_response = compute(prompt, model)
    return similarity.compute(response, expected_response)


def compute_similarity_against_model(
    row: pandas.core.series.Series,
    prompt_column_name: str,
    model: str = "gpt-4",
    response_column_name: str = "response",
) -> str:
    r"""
    Computes the similarity of a given response to the expected result
    generated from a high quality LLM (by default GPT-4) using the same prompt.

    Args:
        row (pandas.core.series.Series): A row of data from the full DataFrame (including input, model response, other
            metrics, etc).
        prompt_column_name (str): name of the column that contains the input prompt
        model (str): name of the model that will serve as the judge
        response_column_name (str): name of the column that contains the model's response, defaults to ``"response"``
    """

    expected_response = compute(row[prompt_column_name], model)
    return similarity.compute(row[response_column_name], expected_response)



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)
import openai
import logging


def generate_retry_decorator(wait_lower_bound: int = 3, wait_upper_bound: int = 12, max_retry_attempts: int = 5):
    r"""
    Creates a retry decorator that can be used for requests. It looks for specific exceptions and waits for
    certain about of time before retrying. This improves the reliability of the request queue.

    Args:
        wait_lower_bound (int): lower bound to the wait time before retry, defaults to 3.
        wait_upper_bound (int): upper bound to the wait time before retry, defaults to 12.
        max_retry_attempts (int): maximum number of retries before stopping, defaults to 5.
    """
    return retry(
        # For the `i`th attempt, wait 2^i seconds before retrying
        # with lower and upper bound of [3s, 12s].
        wait=wait_exponential(multiplier=1, min=wait_lower_bound, max=wait_upper_bound),
        stop=stop_after_attempt(max_retry_attempts),
        reraise=True,
        retry=(  # Retry for these specific exceptions
            retry_if_exception_type(openai.APIConnectionError)
            | retry_if_exception_type(openai.APIError)
            | retry_if_exception_type(openai.RateLimitError)
            | retry_if_exception_type(openai.APIStatusError)
            | retry_if_exception_type(openai.APIConnectionError)
            | retry_if_exception_type(openai.APIResponseValidationError)
            | retry_if_exception_type(openai.APITimeoutError)
        ),
        before_sleep=before_sleep_log(logging.getLogger(__name__), logging.WARNING),
    )


retry_decorator = generate_retry_decorator()



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List, Optional, Type
import jinja2
from .harness import ExperimentationHarness, Experiment
import logging


class PromptTemplateExperimentationHarness(ExperimentationHarness):
    r"""
    An experimentation harness used to test various prompt templates.
    We use `jinja` templates, e.g. "Answer the following question: {{input}}".

    Args:
        experiment (Type[Experiment]): The experiment constructor that you would like to execute within the harness
            (e.g. ``prompttools.experiment.OpenAICompletionExperiment``)
        model_name (str): The name of the model.
        prompt_templates (List[str]): A list of prompt ``jinja``-styled templates.
        user_inputs (List[Dict[str, str]]): A list of dictionaries representing user inputs.
        model_arguments (Optional[Dict[str, object]], optional): Additional arguments for the model.
            Defaults to ``None``.
    """

    PIVOT_COLUMNS = ["prompt_template", "user_input"]

    def __init__(
        self,
        experiment: Type[Experiment],
        model_name: str,
        prompt_templates: List[str],
        user_inputs: List[Dict[str, str]],
        model_arguments: Optional[Dict[str, object]] = None,
    ):
        self.environment = jinja2.Environment()
        self.experiment_cls_constructor = experiment
        self.model_name = model_name
        self.prompt_templates = prompt_templates
        self.user_inputs = user_inputs
        self.model_arguments = {} if model_arguments is None else model_arguments
        super().__init__()

    def prepare(self) -> None:
        r"""
        Creates prompts from templates to use for the experiment, and then initializes and prepares the experiment.
        """
        self.input_pairs_dict = {}
        rendered_inputs = []
        for pt in self.prompt_templates:
            for user_input in self.user_inputs:
                template = self.environment.from_string(pt)
                prompt = template.render(**user_input)
                rendered_inputs.append(prompt)
                self.input_pairs_dict[prompt] = (pt, user_input)
        self.experiment = self.experiment_cls_constructor(
            [self.model_name],
            rendered_inputs,
            **self._prepare_arguments(self.model_arguments),
        )
        super().prepare()

    def run(self):
        if not self.experiment:
            self.prepare()
        super().run()



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


import openai
import prompttools.logger  # noqa: F401 Importing this line will monkey-patch `openai.chat.completions.create`


r"""
Example of using `prompttools.logger`.

All you need to do is call `import prompttools.logger` to start logging.
You can optionally add `hegel_model` to your call (as seen below). This will associate
this call with a specific name in the logs.

The OpenAI call is unchanged, it executes normally between your machine and OpenAI's server.

Note:
You should have "HEGELAI_API_KEY" and "OPENAI_API_KEY" loaded into `os.environ`.
"""

if __name__ == "__main__":
    for i in range(1):
        messages = [
            {"role": "user", "content": f"What is 1 + {i}?"},
        ]

        # `hegel_model` is an optional argument that allows you to tag your call with a specific name
        # Logging still works without this argument
        # The rest of the OpenAI call happens as normal between your machine and OpenAI's server
        openai_response = openai.chat.completions.create(
            model="gpt-3.5-turbo", messages=messages, hegel_model="Math Model"
        )
        print(f"{openai_response = }")

    print("End")



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

try:
    from anthropic import HUMAN_PROMPT, AI_PROMPT
except ImportError:
    HUMAN_PROMPT, AI_PROMPT = None, None


GENERIC_TEMPLATE = """INSTRUCTION:
{instruction}
PROMPT:
{user_input}
RESPONSE:
"""

PALM_TEMPLATE = """{instruction}

{user_input}
"""

LLAMA_TEMPLATE = """<s>[INST] <<SYS>>
{instruction}
<</SYS>
{user_input} [/INST]
"""

ANTHROPIC_TEMPLATE = """{HUMAN_PROMPT}{instruction}
{user_input}
{AI_PROMPT}"""


class PromptSelector:
    r"""
    An abstraction for rendering the same prompt
    for different models, e.g. OpenAI Chat models
    and Llama models
    """

    def __init__(self, instruction: str, user_input: object):
        self.instruction = instruction
        self.user_input = user_input

    def for_openai_chat(self):
        return [
            {"role": "system", "content": self.instruction},
            {"role": "user", "content": self.user_input},
        ]

    def for_openai_completion(self):
        return GENERIC_TEMPLATE.format(instruction=self.instruction, user_input=self.user_input)

    def for_huggingface_hub(self):
        return GENERIC_TEMPLATE.format(instruction=self.instruction, user_input=self.user_input)

    def for_llama(self):
        return LLAMA_TEMPLATE.format(instruction=self.instruction, user_input=self.user_input)

    def for_anthropic(self):
        return ANTHROPIC_TEMPLATE.format(
            HUMAN_PROMPT=HUMAN_PROMPT, instruction=self.instruction, user_input=self.user_input, AI_PROMPT=AI_PROMPT
        )

    def for_palm(self):
        return PALM_TEMPLATE.format(instruction=self.instruction, user_input=self.user_input)



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Callable, Dict, List, Tuple
from queue import Queue, Empty
from time import perf_counter
import threading
import openai
import logging

from prompttools.requests.retries import retry_decorator


class RequestQueue:
    r"""
    A generic queue for processing requests in the `prompttools` library.
    It can be used to handle and time requests to any LLM asynchronously.
    """

    def __init__(self):
        self.data_queue = Queue()
        self.is_running = True
        self.worker_thread = threading.Thread(target=self._process_queue, daemon=True)
        self.worker_thread.start()
        self.request_args: list[dict[str, object]] = []
        self.request_results: list[dict[str, object]] = []
        self.request_latencies: list[float] = []

    def _process_queue(self) -> None:
        while self.is_running:
            try:
                fn, args = self.data_queue.get(timeout=0.2)
                self._do_task(fn, args)
                self.data_queue.task_done()
            except Empty:
                continue

    def _do_task(self, fn: Callable, args: Dict[str, object]) -> None:
        try:
            # TODO: For the streamlit app, we need to set the api key this way.
            # Ideally, OpenAI should be able to use the env var.
            if "OPENAI_API_KEY" in os.environ:
                openai.api_key = os.environ["OPENAI_API_KEY"]
            res = self._run(fn, args)
            self.request_args.append(args)
            self.request_results.append(res[0])
            self.request_latencies.append(res[1])
        # TODO: If we get an unexpected error here, the queue will hang
        except openai.AuthenticationError:
            logging.error("Authentication error. Skipping request.")

    @retry_decorator
    def _run(self, fn: Callable, args: Dict[str, object]) -> Tuple[Dict[str, object], float]:
        start = perf_counter()
        result = fn(**args)
        return result, perf_counter() - start

    def shutdown(self) -> None:
        r"""
        Stops the worker thread from executed and joins it.
        """
        self.data_queue.join()
        self.is_running = False
        # TODO: If we are hanging and interrupt, this line will
        #       have the following error: TypeError: 'NoneType' object is not callable
        self.worker_thread.join()

    def __del__(self) -> None:
        self.shutdown()

    def enqueue(self, callable: Callable, args: Dict[str, object]) -> None:
        r"""
        Adds another request to the queue.
        """
        self.data_queue.put((callable, args))

    def get_input_args(self) -> List[Dict[str, object]]:
        r"""
        Joins the queue and gets input args that lead to the result.
        """
        self.data_queue.join()
        return self.request_args

    def get_results(self) -> List[Dict[str, object]]:
        r"""
        Joins the queue and gets results.
        """
        self.data_queue.join()
        return self.request_results

    def get_latencies(self) -> List[float]:
        r"""
        Joins the queue and gets latencies.
        """
        self.data_queue.join()
        return self.request_latencies



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


from .experiments.experiment import Experiment
from .experiments.openai_chat_experiment import OpenAIChatExperiment
from .experiments.openai_completion_experiment import OpenAICompletionExperiment
from .experiments.anthropic_completion_experiment import AnthropicCompletionExperiment
from .experiments.huggingface_hub_experiment import HuggingFaceHubExperiment
from .experiments.google_palm_experiment import GooglePaLMCompletionExperiment
from .experiments.google_vertex_chat_experiment import GoogleVertexChatCompletionExperiment
from .experiments.llama_cpp_experiment import LlamaCppExperiment
from .experiments.chromadb_experiment import ChromaDBExperiment
from .experiments.weaviate_experiment import WeaviateExperiment
from .experiments.lancedb_experiment import LanceDBExperiment
from .experiments.mindsdb_experiment import MindsDBExperiment
from .experiments.langchain_experiment import SequentialChainExperiment, RouterChainExperiment
from .experiments.stablediffusion_experiment import StableDiffusionExperiment
from .experiments.replicate_experiment import ReplicateExperiment
from .experiments.qdrant_experiment import QdrantExperiment
from .experiments.pinecone_experiment import PineconeExperiment

__all__ = [
    "AnthropicCompletionExperiment",
    "ChromaDBExperiment",
    "Experiment",
    "GooglePaLMCompletionExperiment",
    "GoogleVertexChatCompletionExperiment",
    "LanceDBExperiment",
    "LlamaCppExperiment",
    "HuggingFaceHubExperiment",
    "MindsDBExperiment",
    "OpenAIChatExperiment",
    "OpenAICompletionExperiment",
    "PineconeExperiment",
    "QdrantExperiment",
    "ReplicateExperiment",
    "RouterChainExperiment",
    "SequentialChainExperiment",
    "StableDiffusionExperiment",
    "WeaviateExperiment",
]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Callable, List, Optional
import pandas as pd
import warnings


class Benchmark:
    r"""
    Benchmark models using defined data sets.
    Find example under benchmarks/examples/benchmarking.ipynb.

    Args:
    ----
        experiment (experiment type): experiment to use
        eval_methods (Callable): list of evaluation methods to measure response similarity
        prompts (list(str)): list of queries, questions, prompts for LLMs to respond to
        response_options (list(str)): possible responses to measure against
        correct_response_indices (list(int)): list of index of correct response in response_options
    """

    def __init__(
        self,
        experiment: Any,
        eval_method: Callable,
        prompts: List[str],
        response_options: List[Any],
        correct_response_indices: Optional[List[int]] = None,
    ):
        self.experiment = experiment
        self.eval_method = eval_method
        self.prompts = prompts
        self.response_options = response_options
        self.correct_response_indices = correct_response_indices

    def _get_precision(
        self,
        dataframe: pd.DataFrame,
        pred_col: str,
        label_col: str,
    ) -> float:
        r"""
        Calculate precision.
        """
        # TODO: coming soon
        pass

    def multiple_choice_accuracy(
        self,
        dataframe: pd.DataFrame,
        col1: str,
        col2: str,
    ) -> float:
        r"""
        Benchmark LLM accuracy on multiple choice
        prompt endings.
        """
        correct = 0
        for _, row in dataframe.iterrows():
            if row[col1] == row[col2]:
                correct += 1
        return correct / len(dataframe)

    def multiple_choice_benchmark(
        self,
    ) -> Any:
        r"""
        Run model experiments to measure response quality.
        """
        self.experiment.run()

        if "prompt" not in self.experiment.full_df.columns:
            # Assume messages column is in place of prompt
            self.experiment.full_df["prompt"] = self.experiment.full_df["messages"].map(lambda x: str(x))
            warnings.warn("Column 'prompt' does not exist. Using column 'messages' instead.", UserWarning, stacklevel=2)
        # Get option with highest similarity to LLM response
        benchmark_df = self.experiment.full_df[["prompt", "response"]]
        benchmark_df["response_options"] = self.response_options
        benchmark_df = benchmark_df.explode(column="response_options").reset_index()
        scores = []
        for _, row in benchmark_df.iterrows():
            scores.append(self.eval_method(row=row, expected=row["response_options"]))
        benchmark_df["scores"] = scores
        benchmark_df["max_value"] = benchmark_df.groupby("prompt")["scores"].transform("max")
        benchmark_df = benchmark_df[benchmark_df["scores"] == benchmark_df["max_value"]]
        benchmark_df = benchmark_df.sort_index()
        # Colect model choices
        model_choice = []
        for i, choice in enumerate(benchmark_df["response_options"].values):
            model_choice.append(self.response_options[i].index(choice))
        benchmark_df["model_choice"] = model_choice
        benchmark_df["labels"] = self.correct_response_indices
        return self.multiple_choice_accuracy(benchmark_df, "model_choice", "labels")



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from .benchmark import Benchmark


__all__ = [
    "Benchmark",
]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Any, Dict, List, Union
import itertools

try:
    from huggingface_hub.inference_api import InferenceApi
except ImportError:
    InferenceApi = None

from time import perf_counter
import logging

from prompttools.selector.prompt_selector import PromptSelector
from prompttools.mock.mock import mock_hf_completion_fn

from .experiment import Experiment
from .error import PromptExperimentException

VALID_TASKS = ("text2text-generation", "text-generation", "summarization")


class HuggingFaceHubExperiment(Experiment):
    r"""
    Experiment for Hugging Face Hub's API.
    It accepts lists for each argument passed into Hugging Face Hub's API,
    then creates a cartesian product of those arguments, and gets results for each.

    Note:
        - All arguments here should be a ``list``, even if you want to keep the argument frozen
          (i.e. ``temperature=[1.0]``), because the experiment will try all possible combination
          of the input arguments. For example, ``kwargs`` should have string keys,
          with ``list``s being the values.

    Args:
        repo_id (List[str]): IDs of repository (e.g. [`user/bert-base-uncased`]).
        prompt (List[str] | List[PromptSelector]): list of prompts to test
        task (List[str]): List of tasks in strings. Determines whether to force a task instead of using task
            specified in the repository.
        **kwargs (Dict[str, list[object]]): Keyword parameters used in the call to ``InferenceApi``.
            The values should be ``list``s.
    """

    MODEL_PARAMETERS = ["repo_id", "task"]

    CALL_PARAMETERS = ["prompt"]

    def __init__(
        self,
        repo_id: List[str],
        prompt: Union[List[str], List[PromptSelector]],
        task: List[str] = ["text-generation"],
        **kwargs: Dict[str, list[object]],
    ):
        if InferenceApi is None:
            raise ModuleNotFoundError(
                "Package `huggingface_hub` is required to be installed to use this experiment."
                "Please use `pip install huggingface_hub` to install the package"
            )
        self.completion_fn = self.hf_completion_fn
        if os.getenv("DEBUG", default=False):
            self.completion_fn = mock_hf_completion_fn
        self.model_params = dict(repo_id=repo_id, task=task)

        # If we are using a prompt selector, we need to render
        # messages, as well as create prompt_keys to map the messages
        # to corresponding prompts in other models.
        if isinstance(prompt[0], PromptSelector):
            self.prompt_keys = {selector.for_huggingface_hub(): selector.for_huggingface_hub() for selector in prompt}
            prompt = [selector.for_huggingface_hub() for selector in prompt]
        else:
            self.prompt_keys = prompt

        self.call_params = dict(prompt=prompt)
        for k, v in kwargs.items():
            self.CALL_PARAMETERS.append(k)
            self.call_params[k] = v

        self.all_args = self.model_params | self.call_params
        super().__init__()

    def prepare(self) -> None:
        r"""
        Creates argument combinations by taking the cartesian product of all inputs.
        """
        self.model_argument_combos = [
            dict(zip(self.model_params, val)) for val in itertools.product(*self.model_params.values())
        ]
        self.call_argument_combos = [
            dict(zip(self.call_params, val)) for val in itertools.product(*self.call_params.values())
        ]

    def hf_completion_fn(
        self,
        **params: Dict[str, Any],
    ):
        r"""
        Local model helper function to make request
        """
        client = params["client"]
        model_kwargs = {k: params[k] for k in params if k not in ["client", "prompt"]}
        response = client(inputs=params["prompt"], params=model_kwargs)
        return response

    def run(
        self,
        runs: int = 1,
    ) -> None:
        r"""
        Create tuples of input and output for every possible combination of arguments.
        For each combination, it will execute `runs` times, default to 1.
        # TODO This can be done with an async queue
        """
        if not self.argument_combos:
            logging.info("Preparing first...")
            self.prepare()
        results = []
        latencies = []
        for model_combo in self.model_argument_combos:
            client = InferenceApi(
                repo_id=model_combo["repo_id"],
                token=os.environ.get("HUGGINGFACEHUB_API_TOKEN"),
                task=model_combo["task"],
            )
            for call_combo in self.call_argument_combos:
                for _ in range(runs):
                    call_combo["client"] = client
                    start = perf_counter()
                    res = self.completion_fn(**call_combo)
                    latencies.append(perf_counter() - start)
                    results.append(res)
                    self.argument_combos.append(model_combo | call_combo)
        if len(results) == 0:
            logging.error("No results. Something went wrong.")
            raise PromptExperimentException
        self._construct_result_dfs(self.argument_combos, results, latencies, extract_response_equal_full_result=True)

    @staticmethod
    def _extract_responses(output: List[Dict[str, object]]) -> list[str]:
        return [choice["generated_text"] for choice in output][0]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

r"""
Use a list to optionally hold a reference to the embedding model and client,
allowing for lazy initialization.
"""
from typing import Dict
import pandas.core.series
import logging

try:
    import cv2
except ImportError:
    cv2 = None
try:
    from skimage.metrics import structural_similarity as skimage_structural_similarity
except ImportError:
    skimage_structural_similarity = None


EMBEDDING_MODEL = []
CHROMA_CLIENT = []


def _get_embedding_model():
    if len(EMBEDDING_MODEL) == 0:
        from sentence_transformers import SentenceTransformer

        EMBEDDING_MODEL.append(SentenceTransformer("sentence-transformers/all-MiniLM-L6-v2"))
    return EMBEDDING_MODEL[0]


def _get_chroma_client():
    if len(CHROMA_CLIENT) == 0:
        import chromadb

        CHROMA_CLIENT.append(chromadb.Client())
    return CHROMA_CLIENT[0]


def _from_huggingface(doc1, doc2):
    model = _get_embedding_model()
    embedding_1 = model.encode(doc1, convert_to_tensor=True)
    embedding_2 = model.encode(doc2, convert_to_tensor=True)
    from sentence_transformers.util import pytorch_cos_sim

    return pytorch_cos_sim(embedding_1, embedding_2).item()


def _from_chroma(doc1, doc2):
    chroma_client = _get_chroma_client()
    collection = chroma_client.create_collection(name="test_collection")
    collection.add(documents=[doc1], ids=["id1"])
    query_results = collection.query(query_texts=doc2, n_results=1)
    chroma_client.delete_collection("test_collection")
    return query_results["distances"][0][0] / 2


def compute(doc1, doc2, use_chroma=False):
    r"""
    Computes the semantic similarity between two documents, using either ChromaDB
    or HuggingFace sentence_transformers.

    Args:
        doc1 (str): The first document.
        doc2 (str): The second document.
        use_chroma (bool): Indicates whether or not to use Chroma.
            If ``False``, uses HuggingFace ``sentence_transformers``.
    """
    if use_chroma:
        return _from_chroma(doc1, doc2)
    else:
        return _from_huggingface(doc1, doc2)


def evaluate(prompt: str, response: str, metadata: Dict, expected: str) -> float:
    r"""
    A simple test that checks semantic similarity between the expected response (provided by the user)
    and the model's text responses.

    Args:
        prompt (str): Not used.
        response (str): the response string that will be compared against
        metadata (dict): Not used.
        expected (str): the expected response
    """
    return compute(expected, response)


def structural_similarity(
    row: pandas.core.series.Series, expected: str, response_column_name: str = "response"
) -> float:
    r"""
    Compute the structural similarity index measure (SSIM) between two images.

    Args:
        row (pandas.core.series.Series): A row of data from the full DataFrame (including input, model response, other
            metrics, etc).
        expected (str): the column name of the expected image responses in each row
        response_column_name (str): the column name that contains the model's response, defaults to ``"response"``
    """
    if cv2 is None:
        raise ModuleNotFoundError(
            "Package `cv2` is required to be installed to use this evaluation method."
            "Please use `pip install opencv-python` to install the package"
        )
    if skimage_structural_similarity is None:
        raise ModuleNotFoundError(
            "Package `skimage` is required to be installed to use this evaluation method."
            "Please use `pip install scikit-image` to install the package"
        )
    if len(expected) == 1:
        logging.warning("Expected should be a list of strings. You may have passed in a single string.")
    expected_img = cv2.imread(expected)
    expected_img = cv2.cvtColor(expected_img, cv2.COLOR_BGR2GRAY)
    score, _ = skimage_structural_similarity(row[response_column_name], expected_img, full=True)
    return score


def semantic_similarity(row: pandas.core.series.Series, expected: str, response_column_name: str = "response") -> float:
    r"""
    A simple test that checks semantic similarity between the expected response (provided by the user)
    and the model's text responses.

    Args:
        row (pandas.core.series.Series): A row of data from the full DataFrame (including input, model response, other
            metrics, etc).
        expected (str): the expected responses for each row in the column
        response_column_name (str): name of the column that contains the model's response, defaults to ``"response"``
    """
    if len(expected) == 1:
        logging.warn("Expected should be a list of strings." + "You may have passed in a single string")
    return compute(expected, row[response_column_name])



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Any, Dict, List
import itertools

from time import perf_counter
import logging

try:
    import cv2
except ImportError:
    cv2 = None

try:
    from diffusers import DiffusionPipeline, StableDiffusionPipeline
except ImportError:
    DiffusionPipeline = None

from prompttools.mock.mock import mock_stable_diffusion

from .experiment import Experiment
from .error import PromptExperimentException


class StableDiffusionExperiment(Experiment):
    r"""
    Experiment for experiment with the Stable Diffusion model.

    Args:
        hf_model_path (str): path to model on hugging face
        use_auth_token (bool): boolean to determine if hf login is necessary [needed without GPU]
        kwargs (dict): keyword arguments to call the model with
    """

    MODEL_PARAMETERS = ["hf_model_path", "prompt"]

    CALL_PARAMETERS: List[str] = ["prompt"]

    def __init__(
        self,
        hf_model_path: List[str],
        prompt: List[str],
        compare_images_folder: str,
        use_auth_token: bool = False,
        **kwargs: Dict[str, object],
    ):
        if DiffusionPipeline is None:
            raise ModuleNotFoundError(
                "Package `diffusers` is required to be installed to use this experiment."
                "Please use `pip install diffusers` and \
                `pip install invisible_watermark transformers accelerate safetensors` to install the package"
            )
        if cv2 is None:
            raise ModuleNotFoundError(
                "Package `cv2` is required to be installed to use this experiment."
                "Please use `pip opencv-python` to install the package"
            )
        self.use_auth_token = use_auth_token
        self.completion_fn = self.sd_completion_fn
        self.compare_images_folder = compare_images_folder
        if os.getenv("DEBUG", default=False):
            self.completion_fn = mock_stable_diffusion
        self.model_params = dict(hf_model_path=hf_model_path)

        self.call_params = dict(prompt=prompt)
        for k, v in kwargs.items():
            self.CALL_PARAMETERS.append(k)
            self.call_params[k] = v

        self.all_args = self.model_params | self.call_params
        super().__init__()

    def prepare(self) -> None:
        r"""
        Combo builder.
        """
        self.model_argument_combos = [
            dict(zip(self.model_params, val, strict=False)) for val in itertools.product(*self.model_params.values())
        ]
        self.call_argument_combos = [
            dict(zip(self.call_params, val, strict=False)) for val in itertools.product(*self.call_params.values())
        ]

    def sd_completion_fn(
        self,
        **params: Dict[str, Any],
    ):
        r"""
        Local model helper function to make request.
        """
        client = params["client"]
        image_folder = params["image_folder"]
        if not os.path.exists(image_folder):
            os.makedirs(image_folder)
        response = client(params["prompt"])
        img = response["images"][0]
        img_path = params["image_folder"] + "_".join(params["prompt"].split(" ")) + ".png"
        img.save(img_path)
        main_img = cv2.imread(img_path)
        logging.info("Resizing comparison images to match Stable Diffusion response image size for comparison.")
        for fil in os.listdir(self.compare_images_folder):
            compare_img = cv2.imread(self.compare_images_folder + fil)
            compare_img = cv2.resize(compare_img, (main_img.shape[1], main_img.shape[0]))
            cv2.imwrite(self.compare_images_folder + fil, compare_img)
        return main_img

    def run(
        self,
        runs: int = 1,
    ) -> None:
        r"""
        Create tuples of input and output for every possible combination of arguments.
        For each combination, it will execute `runs` times, default to 1.
        # TODO This can be done with an async queue.
        """
        if not self.argument_combos:
            logging.info("Preparing first...")
            self.prepare()
        results = []
        latencies = []
        for model_combo in self.model_argument_combos:
            for call_combo in self.call_argument_combos:
                if self.use_auth_token:
                    client = StableDiffusionPipeline.from_pretrained(
                        model_combo["hf_model_path"], use_auth_token=self.use_auth_token
                    )
                else:
                    client = DiffusionPipeline.from_pretrained(
                        model_combo["hf_model_path"], **{k: call_combo[k] for k in call_combo if k != "prompt"}
                    )
                    client.to("cuda")
                for _ in range(runs):
                    call_combo["client"] = client
                    start = perf_counter()
                    res = self.completion_fn(**call_combo)
                    latencies.append(perf_counter() - start)
                    results.append(res)
                    self.argument_combos.append(model_combo | call_combo)
        if len(results) == 0:
            logging.error("No results. Something went wrong.")
            raise PromptExperimentException
        self._construct_result_dfs(self.argument_combos, results, latencies, extract_response_equal_full_result=True)

    @staticmethod
    def _extract_responses(output: object) -> object:
        return cv2.cvtColor(output, cv2.COLOR_BGR2GRAY)



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Dict, List, Optional, Union
import openai

from prompttools.selector.prompt_selector import PromptSelector
from prompttools.mock.mock import mock_openai_completion_fn
from .experiment import Experiment


class OpenAICompletionExperiment(Experiment):
    r"""
    This class defines an experiment for OpenAI's completion API.
    It accepts lists for each argument passed into OpenAI's API, then creates
    a cartesian product of those arguments, and gets results for each.

    Note:
        - All arguments here should be a ``list``, even if you want to keep the argument frozen
          (i.e. ``temperature=[1.0]``), because the experiment will try all possible combination
          of the input arguments.
        - For detailed description of the input arguments, please reference at OpenAI's completion API.

    Args:
        model (list[str]): list of ID(s) of the model(s) to use, e.g. ``["gpt-3.5-turbo", "ft:gpt-3.5-turbo:org_id"]``
            If you are using Azure OpenAI service, put the models' deployment names here

        prompt (list[str]): the prompt(s) to generate completions for, encoded as a string, array of strings,
            array of tokens, or array of token arrays.

        suffix: (list[str]): Defaults to [None]. the suffix(es) that comes after a completion of inserted text.

        max_tokens (list[int]):
            Defaults to [inf]. The maximum number of tokens to generate in the chat completion.

        temperature (list[float]):
            Defaults to [1.0]. What sampling temperature to use, between 0 and 2. Higher values like 0.8 will make
            the output more random, while lower values like 0.2 will make it more focused and deterministic.

        top_p (list[float]):
            Defaults to [1.0]. An alternative to sampling with temperature, called nucleus sampling, where the
            model considers the results of the tokens with top_p probability mass. So 0.1 means only the tokens
            comprising the top 10% probability mass are considered.

        n (list[int]):
            Defaults to [1]. How many chat completion choices to generate for each input message.

        stream (list[bool]):
            Defaults to [False]. If set, partial message deltas will be sent, like in ChatGPT. Tokens will be sent
            as data-only server-sent events as they become available, with the stream terminated by a data: [DONE]
            message.

        logprobs (list[int]):
            Defaults to [None]. Include the log probabilities on the ``logprobs`` most likely tokens, as well the
            chosen tokens.

        echo (list[bool]):
            Echo back the prompt in addition to the completion.

        stop (list[list[str]]):
            Defaults to [None]. Up to 4 sequences where the API will stop generating further tokens.

        presence_penalty (list[float]):
            Defaults to [0.0]. Number between -2.0 and 2.0. Positive values penalize new tokens based on whether
            they appear in the text so far, increasing the model's likelihood to talk about new topics.

        frequency_penalty (list[float]):
            Defaults to [0.0]. Number between -2.0 and 2.0. Positive values penalize new tokens based on their
            existing frequency in the text so far, decreasing the model's likelihood to repeat the same line
            verbatim.

        best_of (list[int]):
            Generates ``best_of`` completions server-side and returns the "best" (the one with the highest log
            probability per token). Results cannot be streamed.

        logit_bias (list[dict]):
            Defaults to ``[None]``. Modify the likelihood of specified tokens appearing in the completion. Accepts a
            json object that maps tokens (specified by their token ID in the tokenizer) to an associated bias value
            from -100 to 100.

        azure_openai_service_configs (Optional[dict]):
            Defaults to ``None``. If it is set, the experiment will use Azure OpenAI Service. The input dict should
            contain these 3 keys (but with values based on your use case and configuration):
            ``{"AZURE_OPENAI_ENDPOINT": "https://YOUR_RESOURCE_NAME.openai.azure.com/",
               "API_TYPE": "azure", "API_VERSION": "2023-05-15"``
    """

    def __init__(
        self,
        model: List[str],
        prompt: Union[List[str], List[PromptSelector]],
        suffix: Optional[List[str]] = [None],
        max_tokens: Optional[List[int]] = [float("inf")],
        temperature: Optional[List[float]] = [1.0],
        top_p: Optional[List[float]] = [1.0],
        n: Optional[List[int]] = [1],
        stream: Optional[List[bool]] = [False],
        logprobs: Optional[List[int]] = [None],
        echo: Optional[List[bool]] = [False],
        stop: Optional[List[List[str]]] = [None],
        presence_penalty: Optional[List[float]] = [0],
        frequency_penalty: Optional[List[float]] = [0],
        best_of: Optional[List[int]] = [1],
        logit_bias: Optional[Dict] = [None],
        azure_openai_service_configs: Optional[dict] = None,
    ):
        self.completion_fn = openai.completions.create
        if os.getenv("DEBUG", default=False):
            self.completion_fn = mock_openai_completion_fn

        # If we are using a prompt selector, we need to render
        # messages, as well as create prompt_keys to map the messages
        # to corresponding prompts in other models.
        if isinstance(prompt[0], PromptSelector):
            self.prompt_keys = {
                selector.for_openai_completion(): selector.for_openai_completion() for selector in prompt
            }
            prompt = [selector.for_openai_completion() for selector in prompt]
        else:
            self.prompt_keys = prompt

        self.all_args = dict(
            model=model,
            prompt=prompt,
            suffix=suffix,
            max_tokens=max_tokens,
            temperature=temperature,
            top_p=top_p,
            n=n,
            stream=stream,
            logprobs=logprobs,
            echo=echo,
            stop=stop,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            best_of=best_of,
            logit_bias=logit_bias,
        )

        # These parameters aren't supported by `gpt-35-turbo`, we can remove them if they are equal to defaults
        # This has no impact on the default case
        if self.all_args["echo"] == [False]:
            del self.all_args["echo"]
        if self.all_args["logit_bias"] == [None]:
            del self.all_args["logit_bias"]
        if self.all_args["best_of"] == [1]:
            del self.all_args["best_of"]

        if azure_openai_service_configs:
            openai.api_key = os.environ["AZURE_OPENAI_KEY"]
            openai.api_base = azure_openai_service_configs["AZURE_OPENAI_ENDPOINT"]
            openai.api_type = azure_openai_service_configs["API_TYPE"]
            openai.api_version = azure_openai_service_configs["API_VERSION"]
            del self.all_args["model"]
            self.all_args["engine"] = model

        super().__init__()

    @staticmethod
    def _extract_responses(output: openai.types.Completion) -> list[str]:
        return [choice.text for choice in output.choices][0]

    def _get_model_names(self):
        return [combo["model"] for combo in self.argument_combos]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


import os
from os.path import join, dirname

try:
    from dotenv import load_dotenv
except ImportError:
    load_dotenv = None

if load_dotenv is not None:
    dotenv_path = join(dirname(dirname(__file__)), ".env")
    load_dotenv(dotenv_path)


ENV = os.environ.get("ENV", "prod")
if ENV == "development":
    HEGEL_BACKEND_URL = """http://127.0.0.1:5000"""
else:
    HEGEL_BACKEND_URL = """https://api.hegel-ai.com"""



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


from prompttools.experiment import LlamaCppExperiment
from prompttools.experiment import OpenAIChatExperiment
from prompttools.experiment import OpenAICompletionExperiment
from prompttools.experiment import AnthropicCompletionExperiment
from prompttools.experiment import GooglePaLMCompletionExperiment
from prompttools.experiment import HuggingFaceHubExperiment
from prompttools.experiment import ReplicateExperiment

ENVIRONMENT_VARIABLE = {
    "Replicate": "REPLICATE_API_TOKEN",
    "OpenAI Chat": "OPENAI_API_KEY",
    "OpenAI Completion": "OPENAI_API_KEY",
    "Anthropic": "ANTHROPIC_API_KEY",
    "Google PaLM": "GOOGLE_PALM_API_KEY",
    "HuggingFace Hub": "HUGGINGFACEHUB_API_TOKEN",
}

EXPERIMENTS = {
    "LlamaCpp Chat": LlamaCppExperiment,
    "OpenAI Chat": OpenAIChatExperiment,
    "OpenAI Completion": OpenAICompletionExperiment,
    "Anthropic": AnthropicCompletionExperiment,
    "Google PaLM": GooglePaLMCompletionExperiment,
    "HuggingFace Hub": HuggingFaceHubExperiment,
    "Replicate": ReplicateExperiment,
}

MODES = ("Instruction", "Prompt Template", "Model Comparison")

MODEL_TYPES = (
    "OpenAI Chat",
    "OpenAI Completion",
    "Anthropic",
    "Google PaLM",
    "LlamaCpp Chat",
    "LlamaCpp Completion",
    "HuggingFace Hub",
    "Replicate",
)

OPENAI_CHAT_MODELS = (
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-16k",
    "gpt-3.5-turbo-0613",
    "gpt-3.5-turbo-16k-0613",
    "gpt-3.5-turbo-0301",
    "gpt-4",
    "gpt-4-0613",
    "gpt-4-32k",
    "gpt-4-32k-0613",
    "gpt-4-0314",
    "gpt-4-32k-0314",
)

OPENAI_COMPLETION_MODELS = ("text-davinci-003", "text-davinci-002", "code-davinci-002")



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List, Optional, Callable

import pandas as pd

from .harness import ExperimentationHarness
from prompttools.experiment import OpenAIChatExperiment
from .utility import is_interactive
from IPython import display
from tabulate import tabulate
import logging
from copy import deepcopy


class ModelComparisonHarness(ExperimentationHarness):
    r"""
    An experimentation harness used for comparing models.

    Args:
        model_names (List[str]): The names of the models that you would like to compare
        system_prompts (List[str]): A list of system messages, one for each model.
        model_arguments (List[Optional[Dict]]): A list of model arguments, one for each model.
        user_messages (List[str]) User messages that will be tested across models. Defaults to ``[]``.
        runs (int): Number of runs to execute. Defaults to ``1``.
    """

    _experiment_type = "Comparison"
    PIVOT_COLUMNS = ["model", "messages"]

    def __init__(
        self,
        model_names: List[str],
        system_prompts: List[str],
        user_messages: List[str],
        model_arguments: List[Optional[dict]] = [],
        runs: int = 1,
    ):
        self.model_names = model_names
        self.system_prompts = system_prompts
        self.model_arguments = deepcopy(model_arguments)
        self.user_messages = user_messages
        self.runs = runs
        self.experiments = []

        if len(model_names) != len(system_prompts):
            raise RuntimeError(
                "The number of models must match the number of system messages, because each"
                "system prompt correspond to one model."
            )
        if model_arguments != [] and len(model_arguments) != len(model_names):
            raise RuntimeError(
                "The number of models must match the number of model argument dictionaries,"
                "because each dictionary of arguments correspond to one model."
            )
        super().__init__()

    def prepare(self) -> None:
        """
        Initializes and prepares the experiment.
        """
        self.experiments = []
        for i, model in enumerate(self.model_names):
            system_prompt = self.system_prompts[i]
            model_args = {} if self.model_arguments == [] else self.model_arguments[i]
            messages = []
            for message in self.user_messages:
                messages.append(
                    [
                        self._create_system_prompt(system_prompt),
                        self._create_human_message(message),
                    ]
                )

            experiment = OpenAIChatExperiment(
                [model],
                messages,
                **self._prepare_arguments(model_args),
            )
            self.experiments.append(experiment)

    @staticmethod
    def _create_system_prompt(content: str) -> Dict[str, str]:
        return {"role": "system", "content": content}

    @staticmethod
    def _create_human_message(content: str) -> Dict[str, str]:
        return {"role": "user", "content": content}

    @property
    def full_df(self):
        return self._full_df

    @property
    def partial_df(self):
        return self._partial_df

    @property
    def score_df(self):
        return self._score_df

    def run(self, clear_previous_results: bool = False):
        if not self.experiments:
            self.prepare()
        for exp in self.experiments:
            exp.run(clear_previous_results=clear_previous_results)
        self._update_dfs()

    def evaluate(self, metric_name: str, eval_fn: Callable, static_eval_fn_kwargs: dict = {}, **eval_fn_kwargs) -> None:
        r"""
        Uses the given eval_fn to evaluate the results of the underlying experiment.
        """
        if not self.experiments:
            raise RuntimeError("Cannot evaluate experiment without running first.")
        for exp in self.experiments:
            exp.evaluate(metric_name, eval_fn, static_eval_fn_kwargs, **eval_fn_kwargs)
        self._update_dfs()

    def get_table(self, get_all_cols: bool = False) -> pd.DataFrame:
        columns_to_hide = [
            "stream",
            "response_id",
            "response_choices",
            "response_created",
            "response_created",
            "response_object",
            "response_model",
            "response_system_fingerprint",
            "revision_id",
            "log_id",
        ]

        if get_all_cols:
            return self.full_df
        else:
            table = self.full_df
            columns_to_hide.extend(
                [
                    col
                    for col in ["temperature", "top_p", "n", "presence_penalty", "frequency_penalty"]
                    if table[col].nunique() == 1  # Note this is checking for uniqueness
                ]
            )
            for col in columns_to_hide:
                if col in table.columns:
                    table = table.drop(col, axis=1)
            return table

    def _update_dfs(self):
        self._full_df = pd.concat([exp.full_df for exp in self.experiments], axis=0, ignore_index=True)
        self._partial_df = pd.concat([exp.partial_df for exp in self.experiments], axis=0, ignore_index=True)
        self._score_df = pd.concat([exp.score_df for exp in self.experiments], axis=0, ignore_index=True)

    def visualize(self, get_all_cols: bool = False):
        table = self.get_table(get_all_cols)
        if is_interactive():
            display.display(table)
        else:
            logging.getLogger().setLevel(logging.INFO)
            logging.info(tabulate(table, headers="keys", tablefmt="psql"))

    def _get_state(self):
        state_params = {
            "model_names": self.model_names,
            "system_prompts": self.system_prompts,
            "user_messages": self.user_messages,
            "model_arguments": self.model_arguments,
            "child_experiment_states": [exp._get_state() for exp in self.experiments] if self.experiments else [],
        }
        state = (
            state_params,
            self.full_df,
        )
        print("Creating state of experiment...")
        return state

    @classmethod
    def _load_state(cls, state, experiment_id: str, revision_id: str, experiment_type_str: str):
        (
            state_params,
            full_df,
        ) = state
        if experiment_type_str != cls._experiment_type:
            raise RuntimeError(
                f"The Experiment Type you are trying to load is {experiment_type_str},"
                "which does not match the current class."
            )

        model_names = state_params["model_names"]
        system_prompts = state_params["system_prompts"]
        user_messages = state_params["user_messages"]
        model_arguments = state_params["model_arguments"]
        child_experiment_states = state_params["child_experiment_states"]

        harness = cls(model_names, system_prompts, user_messages, model_arguments)
        harness.experiments = [
            OpenAIChatExperiment._load_state(state, None, None, OpenAIChatExperiment._experiment_type)
            for state in child_experiment_states
        ]
        harness._update_dfs()
        harness._experiment_id = experiment_id
        harness._revision_id = revision_id
        print("Loaded harness.")
        return harness
